#ifndef SWISSPARSE_H
#define SWISSPARSE_H

#include "swissdef.h"

inline void *swiss_mallocretrmsg();
inline void *swiss_mallocstoremsg(struct account *acc);
char *sw_parse_cred(struct account *const acc, char *buf);
char *sw_marsh_cred(struct account *const acc, char *buf);
char *sw_parse_dsize(struct account *const acc, char *buf);
char *sw_parse_data(struct account *const acc, char *buf);
char *sw_marsh_dsize_data(struct account *acc, char *buf);


#endif
